"""

@kylel, benjaminn

"""

import json
import pathlib
import unittest

from papermage.magelib import Document, Entity, Span
from papermage.parsers import PDFPlumberParser
from papermage.predictors import HFBIOTaggerPredictor

TEST_SCIBERT_WEIGHTS = "allenai/scibert_scivocab_uncased"


class TestEntityClassificationPredictor(unittest.TestCase):
    def setUp(self):
        self.fixture_path = pathlib.Path(__file__).parent.parent / "fixtures"
        with open(self.fixture_path / "entity_classification_predictor_test_doc_papermage.json", "r") as f:
            test_doc_json = json.load(f)
        self.doc = Document.from_json(doc_json=test_doc_json)
        ent1 = Entity(spans=[Span(start=86, end=456)])
        ent2 = Entity(spans=[Span(start=457, end=641)])
        self.doc.annotate_entity(field_name="bibs", entities=[ent1, ent2])

        self.predictor = HFBIOTaggerPredictor.from_pretrained(
            model_name_or_path=TEST_SCIBERT_WEIGHTS, entity_name="tokens", context_name="pages"
        )

    def test_predict_pages_tokens(self):
        predictor = HFBIOTaggerPredictor.from_pretrained(
            model_name_or_path=TEST_SCIBERT_WEIGHTS, entity_name="tokens", context_name="pages"
        )
        token_tags = predictor.predict(doc=self.doc)
        assert len(token_tags) == 1

        self.doc.annotate_entity(field_name="token_tags", entities=token_tags)
        for token_tag in token_tags:
            assert isinstance(token_tag.metadata.label, str)
            assert isinstance(token_tag.metadata.score, float)

    def test_predict_bibs_tokens(self):
        self.predictor.context_name = "bibs"
        token_tags = self.predictor.predict(doc=self.doc)
        assert len(token_tags) == 1

    def test_missing_fields(self):
        self.predictor.entity_name = "OHNO"
        with self.assertRaises(AssertionError) as e:
            self.predictor.predict(doc=self.doc)
            assert "OHNO" in e.exception

        self.predictor.entity_name = "tokens"
        self.predictor.context_name = "BLABLA"
        with self.assertRaises(AssertionError) as e:
            self.predictor.predict(doc=self.doc)
            assert "BLABLA" in e.exception

        self.predictor.context_name = "pages"

    def test_predict_pages_tokens_roberta(self):
        predictor = HFBIOTaggerPredictor.from_pretrained(
            model_name_or_path="roberta-base",
            entity_name="tokens",
            context_name="pages",
            add_prefix_space=True,  # Needed for roberta
        )
        token_tags = predictor.predict(doc=self.doc)
        assert len(token_tags) == 1

        self.doc.annotate_entity(field_name="token_tags", entities=token_tags)
        for token_tag in token_tags:
            assert isinstance(token_tag.metadata.label, str)
            assert isinstance(token_tag.metadata.score, float)



"""


@kylel

"""

import unittest

from papermage.magelib import Document, Entity, Span
from papermage.predictors import WhitespacePredictor


class TestWhitespacePredictor(unittest.TestCase):
    def test_predict(self):
        # fmt:off
        #          0         10        20        30        40        50        60        70
        #          01234567890123456789012345678901234567890123456789012345678901234567890123456789
        symbols = "The goal of meta-learning is to train a model on a vari&ety of learning tasks"
        # fmt:on

        spans = [
            Span(start=0, end=3),  # The
            Span(start=4, end=8),  # goal
            Span(start=9, end=11),  # of
            Span(start=12, end=16),  # meta
            Span(start=16, end=17),  # -
            Span(start=17, end=25),  # learning
            Span(start=26, end=28),  # is
            Span(start=29, end=31),  # to
            Span(start=32, end=37),  # train
            Span(start=38, end=39),  # a
            Span(start=40, end=45),  # model
            Span(start=46, end=48),  # on
            Span(start=49, end=50),  # a
            Span(start=51, end=56),  # vari&
            Span(start=56, end=59),  # ety
            Span(start=60, end=62),  # of
            Span(start=63, end=71),  # learning
            Span(start=72, end=77),  # tasks
        ]

        doc = Document(symbols=symbols)
        doc.annotate_entity(field_name="tokens", entities=[Entity(spans=[span]) for i, span in enumerate(spans)])

        predictor = WhitespacePredictor()
        ws_chunks = predictor.predict(doc)

        doc.annotate_entity(field_name="ws_chunks", entities=ws_chunks)
        self.assertEqual(
            [c.text for c in doc.ws_chunks],
            [
                "The",
                "goal",
                "of",
                "meta-learning",
                "is",
                "to",
                "train",
                "a",
                "model",
                "on",
                "a",
                "vari&ety",
                "of",
                "learning",
                "tasks",
            ],
        )
        self.assertEqual(" ".join([c.text for c in doc.ws_chunks]), doc.symbols)



"""

@kylel

"""

import unittest

from papermage.magelib import Annotation


class DummyAnnotation(Annotation):
    pass


class DummyDoc:
    pass


class TestAnnotation(unittest.TestCase):
    def test_abstract_methods(self):
        # trick to test abstract methods from
        # https://clamytoe.github.io/articles/2020/Mar/12/testing-abcs-with-abstract-methods-with-pytest/
        Annotation.__abstractmethods__ = set()
        a = DummyAnnotation()

        to_json = a.to_json()
        from_json = a.from_json(annotation_json={"doesnt_matter": "whats_here"})
        self.assertIsNone(to_json)
        self.assertIsNone(from_json)

    def test_doc(self):
        d = DummyDoc()
        a = Annotation()

        # defaults to None
        self.assertIsNone(a.doc)

        # attaches reference to the Doc object
        a.doc = d
        self.assertIs(a.doc, d)

        # protected setter
        with self.assertRaises(AttributeError) as e:
            a.doc = DummyDoc()

        # detaches from Doc
        a.doc = None
        self.assertIsNone(a.doc)

    def test_id(self):
        a = Annotation()

        # defaults to None
        self.assertIsNone(a.id)

        # setting id doesnt work without Doc
        with self.assertRaises(AttributeError):
            a.id = 12345

        # setting id works w/ a Doc first
        d = DummyDoc()
        a.doc = d
        a.id = 12345
        self.assertEqual(a.id, 12345)



"""

Annotations are objects that are 'aware' of the Document. For example, imagine an entity
in a document; representing it as an Annotation data type would allow you to access the
Document object directly from within the Entity itself.

@kylel

"""

from abc import abstractmethod
from typing import TYPE_CHECKING, Dict, List, Optional, Union

if TYPE_CHECKING:
    from .document import Document


class Annotation:
    """Represent a "unit" (e.g. highlighted span, drawn boxes) layered on a Document."""

    @abstractmethod
    def __init__(self) -> None:
        self._id: Optional[int] = None
        self._doc: Optional["Document"] = None

    @property
    def doc(self) -> Optional["Document"]:
        return self._doc

    @doc.setter
    def doc(self, doc: Optional["Document"]) -> None:
        """This method attaches a Document to this Annotation, allowing the Annotation
        to access things beyond itself within the Document (e.g. neighboring annotations)"""
        if self.doc and doc:
            raise AttributeError(
                "Already has an attached Document. Since Annotations should be"
                "specific to a given Document, we recommend creating a new"
                "Annotation from scratch and then attaching your Document."
            )
        self._doc = doc

    @property
    def id(self) -> Optional[int]:
        return self._id

    @id.setter
    def id(self, id: int) -> None:
        """This method assigns an ID to an Annotation. Requires a Document to be attached
        to this Annotation. ID basically gives the Annotation itself awareness of its
        position within the broader Document."""
        if self.id:
            raise AttributeError(f"This Annotation already has an ID: {self.id}")
        if not self.doc:
            raise AttributeError("This Annotation is missing a Document")
        self._id = id

    @abstractmethod
    def to_json(self) -> Union[Dict, List]:
        pass

    @classmethod
    @abstractmethod
    def from_json(cls, annotation_json: Union[Dict, List]) -> "Annotation":
        pass

    def __getattr__(self, field: str) -> List["Annotation"]:
        """This Overloading is convenient syntax since the `entity.layer` operation is intuitive for folks."""
        try:
            return self.find_by_span(field=field)
        except ValueError:
            # maybe users just want some attribute of the Annotation object
            return self.__getattribute__(field)

    def find_by_span(self, field: str) -> List["Annotation"]:
        """This method allows you to access overlapping Annotations
        within the Document based on Span"""
        if self.doc is None:
            raise ValueError("This annotation is not attached to a document")

        if field in self.doc.fields:
            return self.doc.find_by_span(self, field)
        else:
            raise ValueError(f"Field {field} not found in Document")

    def find_by_box(self, field: str) -> List["Annotation"]:
        """This method allows you to access overlapping Annotations
        within the Document based on Box"""

        if self.doc is None:
            raise ValueError("This annotation is not attached to a document")

        if field in self.doc.fields:
            return self.doc.find_by_box(self, field)
        else:
            raise ValueError(f"Field {field} not found in Document")



"""

@kylel

"""

import unittest

from papermage.magelib import (
    Box,
    Document,
    Entity,
    EntityBoxIndexer,
    EntitySpanIndexer,
    Span,
)


class TestEntitySpanIndexer(unittest.TestCase):
    def test_index_empty(self):
        indexer = EntitySpanIndexer(entities=[])

    def test_overlap_within_single_entity_fails_checks(self):
        entities = [Entity(spans=[Span(0, 5), Span(4, 7)])]

        with self.assertRaises(ValueError):
            EntitySpanIndexer(entities=entities)

    def test_overlap_between_entities_fails_checks(self):
        entities = [Entity(spans=[Span(0, 5), Span(5, 8)]), Entity(spans=[Span(6, 10)])]

        with self.assertRaises(ValueError):
            EntitySpanIndexer(entities)

    def test_finds_matching_entities(self):
        entities_to_index = [
            Entity(spans=[Span(0, 5), Span(5, 8)]),
            Entity(spans=[Span(9, 10)]),
            Entity(spans=[Span(100, 105)]),
        ]

        index = EntitySpanIndexer(entities_to_index)

        # should intersect 1 and 2 but not 3
        probe = Entity(spans=[Span(1, 7), Span(9, 20)])
        matches = index.find(probe)

        self.assertEqual(len(matches), 2)
        self.assertEqual(matches, [entities_to_index[0], entities_to_index[1]])

    def test_finds_matching_entities_in_original_order(self):
        entities_to_index = [
            Entity(spans=[Span(100, 105)]),
            Entity(spans=[Span(9, 10)]),
            Entity(spans=[Span(0, 5), Span(5, 8)]),
        ]

        for i, entity in enumerate(entities_to_index):
            entity.doc = Document(symbols="test")
            entity.id = i

        index = EntitySpanIndexer(entities_to_index)

        # should intersect 2 and 3 but not 1
        probe = Entity(spans=[Span(9, 20), Span(1, 7)])
        matches = index.find(probe)

        self.assertEqual(len(matches), 2)
        self.assertEqual(matches, [entities_to_index[1], entities_to_index[2]])


class TestEntityBoxIndexer(unittest.TestCase):
    def test_overlap_within_single_entity_fails_checks(self):
        entities = [Entity(boxes=[Box(0, 0, 5, 5, page=0), Box(4, 4, 7, 7, page=0)])]

        with self.assertRaises(ValueError):
            EntityBoxIndexer(entities=entities, allow_overlap=False)
        EntityBoxIndexer(entities=entities, allow_overlap=True)

    def test_overlap_between_entities_fails_checks(self):
        entities = [
            Entity(boxes=[Box(0, 0, 5, 5, page=0), Box(5.01, 5.01, 8, 8, page=0)]),
            Entity(boxes=[Box(6, 6, 10, 10, page=0)]),
        ]

        with self.assertRaises(ValueError):
            EntityBoxIndexer(entities=entities, allow_overlap=False)
        EntityBoxIndexer(entities=entities, allow_overlap=True)

    def test_finds_matching_entities_in_doc_order(self):
        entities_to_index = [
            Entity(boxes=[Box(0, 0, 1, 1, page=0), Box(2, 2, 1, 1, page=0)]),
            Entity(boxes=[Box(4, 4, 1, 1, page=0)]),
            Entity(boxes=[Box(100, 100, 1, 1, page=0)]),
        ]
        index = EntityBoxIndexer(entities_to_index)

        # should intersect 1 and 2 but not 3
        probe = Entity(boxes=[Box(1, 1, 5, 5, page=0), Box(9, 9, 5, 5, page=0)])
        matches = index.find(probe)

        self.assertEqual(len(matches), 2)
        self.assertEqual(matches, [entities_to_index[0], entities_to_index[1]])

    def test_finds_matching_entities_accounts_for_pages(self):
        entities_to_index = [
            Entity(boxes=[Box(0.0, 0.0, 0.1, 0.1, page=0), Box(0.2, 0.2, 0.1, 0.1, page=1)]),
            Entity(boxes=[Box(0.4, 0.4, 0.1, 0.1, page=1)]),
            Entity(boxes=[Box(10.0, 10.0, 0.1, 0.1, page=0)]),
        ]

        index = EntityBoxIndexer(entities_to_index)

        # shouldnt intersect any given page 0
        probe = Entity(boxes=[Box(0.1, 0.1, 0.5, 0.5, page=0), Box(0.9, 0.9, 0.5, 0.5, page=0)])
        matches = index.find(probe)

        self.assertEqual(len(matches), 1)
        self.assertEqual(matches, [entities_to_index[0]])

        # shoudl intersect after switching to page 1 (and the page 2 box doesnt intersect)
        probe = Entity(boxes=[Box(0.1, 0.1, 0.5, 0.5, page=1), Box(10.0, 10.0, 0.1, 0.1, page=2)])
        matches = index.find(probe)

        self.assertEqual(len(matches), 2)
        self.assertEqual(matches, [entities_to_index[0], entities_to_index[1]])



"""


@kylel

"""

from itertools import chain
from typing import Dict, List, NamedTuple, Optional, Tuple, Union

from .span import Span
from .box import Box
from .image import Image
from .metadata import Metadata
from .entity import Entity
from .indexer import EntitySpanIndexer, EntityBoxIndexer


# document field names
SymbolsFieldName = "symbols"
ImagesFieldName = "images"
MetadataFieldName = "metadata"
EntitiesFieldName = "entities"
RelationsFieldName = "relations"

PagesFieldName = "pages"
TokensFieldName = "tokens"
RowsFieldName = "rows"
BlocksFieldName = "blocks"
WordsFieldName = "words"
SentencesFieldName = "sentences"
ParagraphsFieldName = "paragraphs"

# these come from vila
TitlesFieldName = "titles"
AuthorsFieldName = "authors"
AbstractsFieldName = "abstracts"
KeywordsFieldName = "keywords"
SectionsFieldName = "sections"
ListsFieldName = "lists"
BibliographiesFieldName = "bibliographies"
EquationsFieldName = "equations"
AlgorithmsFieldName = "algorithms"
FiguresFieldName = "figures"
TablesFieldName = "tables"
CaptionsFieldName = "captions"
HeadersFieldName = "headers"
FootersFieldName = "footers"
FootnotesFieldName = "footnotes"


class Prediction(NamedTuple):
    name: str
    entities: List[Entity]


class Document:
    SPECIAL_FIELDS = [SymbolsFieldName, ImagesFieldName, MetadataFieldName, EntitiesFieldName, RelationsFieldName]

    def __init__(self, symbols: str, metadata: Optional[Metadata] = None):
        self.symbols = symbols
        self.metadata = metadata if metadata else Metadata()
        self.__entity_span_indexers: Dict[str, EntitySpanIndexer] = {}
        self.__entity_box_indexers: Dict[str, EntityBoxIndexer] = {}

    @property
    def fields(self) -> List[str]:
        return list(self.__entity_span_indexers.keys()) + self.SPECIAL_FIELDS

    def find(self, query: Union[Span, Box], field_name: str) -> List[Entity]:
        if isinstance(query, Span):
            return self.__entity_span_indexers[field_name].find(query=Entity(spans=[query]))
        elif isinstance(query, Box):
            return self.__entity_box_indexers[field_name].find(query=Entity(boxes=[query]))
        else:
            raise TypeError(f"Unsupported query type {type(query)}")

    def find_by_span(self, query: Entity, field_name: str) -> List[Entity]:
        # TODO: will rename this to `intersect_by_span`
        return self.__entity_span_indexers[field_name].find(query=query)

    def find_by_box(self, query: Entity, field_name: str) -> List[Entity]:
        # TODO: will rename this to `intersect_by_span`
        return self.__entity_box_indexers[field_name].find(query=query)

    def check_field_name_availability(self, field_name: str) -> None:
        if field_name in self.SPECIAL_FIELDS:
            raise AssertionError(f"{field_name} not allowed Document.SPECIAL_FIELDS.")
        if field_name in self.__entity_span_indexers.keys():
            raise AssertionError(f'{field_name} already exists. Try `doc.remove_entity("{field_name}")` first.')
        if field_name in dir(self):
            raise AssertionError(f"{field_name} clashes with Document class properties.")

    def get_entity(self, field_name: str) -> List[Entity]:
        return getattr(self, field_name)

    def annotate(self, *predictions: Union[Prediction, Tuple[Prediction, ...]]) -> None:
        all_preds = chain.from_iterable([p] if isinstance(p, Prediction) else p for p in predictions)
        for prediction in all_preds:
            self.annotate_entity(field_name=prediction.name, entities=prediction.entities)

    def annotate_entity(self, field_name: str, entities: List[Entity]) -> None:
        self.check_field_name_availability(field_name=field_name)

        for i, entity in enumerate(entities):
            entity.doc = self
            entity.id = i

        self.__entity_span_indexers[field_name] = EntitySpanIndexer(entities=entities)
        self.__entity_box_indexers[field_name] = EntityBoxIndexer(entities=entities)
        setattr(self, field_name, entities)

    def remove_entity(self, field_name: str):
        for entity in getattr(self, field_name):
            entity.doc = None

        delattr(self, field_name)
        del self.__entity_span_indexers[field_name]
        del self.__entity_box_indexers[field_name]

    def get_relation(self, name: str) -> List["Relation"]:
        raise NotImplementedError

    def annotate_relation(self, name: str) -> None:
        self.check_field_name_availability(field_name=name)
        raise NotImplementedError

    def remove_relation(self, name: str) -> None:
        raise NotImplementedError

    def annotate_images(self, images: List[Image]) -> None:
        if len(images) == 0:
            raise ValueError("No images were provided")

        image_types = {type(image) for image in images}
        if len(image_types) > 1:
            raise TypeError(f"Images contain multiple types: {image_types}")
        image_type = image_types.pop()

        if not issubclass(image_type, Image):
            raise NotImplementedError(f"Unsupported image type {image_type} for {ImagesFieldName}")

        setattr(self, ImagesFieldName, images)

    def remove_images(self) -> None:
        raise NotImplementedError

    def to_json(self, field_names: Optional[List[str]] = None, with_images: bool = False) -> Dict:
        """Returns a dictionary that's suitable for serialization

        Use `fields` to specify a subset of groups in the Document to include (e.g. 'sentences')

        Output format looks like
            {
                symbols: "...",
                entities: {...},
                relations: {...},
                metadata: {...}
            }
        """
        # 1) instantiate basic Document dict
        doc_dict = {
            SymbolsFieldName: self.symbols,
            MetadataFieldName: self.metadata.to_json(),
            EntitiesFieldName: {},
            RelationsFieldName: {},
        }

        # 2) serialize each field to JSON
        field_names = list(self.__entity_span_indexers.keys()) if field_names is None else field_names
        for field_name in field_names:
            doc_dict[EntitiesFieldName][field_name] = [entity.to_json() for entity in getattr(self, field_name)]

        # 3) serialize images if `with_images == True`
        if with_images:
            doc_dict[ImagesFieldName] = [image.to_base64() for image in getattr(self, ImagesFieldName)]

        return doc_dict

    @classmethod
    def from_json(cls, doc_json: Dict) -> "Document":
        # 1) instantiate basic Document
        symbols = doc_json[SymbolsFieldName]
        doc = cls(symbols=symbols, metadata=Metadata(**doc_json.get(MetadataFieldName, {})))

        # 2) instantiate entities
        for field_name, entity_jsons in doc_json[EntitiesFieldName].items():
            entities = [Entity.from_json(entity_json=entity_json) for entity_json in entity_jsons]
            doc.annotate_entity(field_name=field_name, entities=entities)

        return doc



"""

Uses a whitespace tokenizer on Document.symbols to predict which `tokens` were originally
part of the same segment/chunk (e.g. "few-shot" if tokenized as ["few", "-", "shot"]).

@kylel

"""

from typing import List, Optional, Set, Tuple

import tokenizers

from papermage.magelib import Document, Entity, Metadata, Span, TokensFieldName
from papermage.predictors.base_predictor import BasePredictor


class WhitespacePredictor(BasePredictor):
    REQUIRED_BACKENDS = None
    REQUIRED_DOCUMENT_FIELDS = [TokensFieldName]

    _dictionary: Optional[Set[str]] = None

    def __init__(self) -> None:
        self.whitespace_tokenizer = tokenizers.pre_tokenizers.WhitespaceSplit()

    def predict(self, doc: Document) -> List[Entity]:
        self._doc_field_checker(doc)

        # 1) whitespace tokenization on symbols. each token is a nested tuple ('text', (start, end))
        ws_tokens: List[Tuple] = self.whitespace_tokenizer.pre_tokenize_str(doc.symbols)

        # 2) filter to just the chunks that are greater than 1 token. Reformat.
        # chunks = []
        # for text, (start, end) in ws_tokens:
        #     overlapping_tokens = document.find_overlapping(
        #         query=SpanGroup(spans=[Span(start=start, end=end)]),
        #         field_name=Tokens
        #     )
        #     if len(overlapping_tokens) > 1:
        #         chunk = SpanGroup(spans=[Span(start=start, end=end)], metadata=Metadata(text=text))
        #         chunks.append(chunk)
        chunks = []
        for text, (start, end) in ws_tokens:
            chunk = Entity(spans=[Span(start=start, end=end)], metadata=Metadata(text=text))
            chunks.append(chunk)
        return chunks



"""

@kylel, benjaminn

"""

from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
import transformers
from smashed.interfaces.simple import (
    FixedBatchSizeMapper,
    FromTokenizerListCollatorMapper,
    Python2TorchMapper,
    TokenizerMapper,
    UnpackingMapper,
)

from papermage.magelib import Annotation, Box, Document, Entity, Metadata, Span
from papermage.predictors.base_predictor import BasePredictor


class BIOBatch:
    def __init__(
        self,
        input_ids: List[List[int]],
        attention_mask: List[List[int]],
        entity_ids: List[List[Optional[int]]],
        context_id: List[int],
    ):
        if not (len(input_ids) == len(attention_mask) == len(entity_ids) == len(context_id)):
            raise ValueError("Inputs to batch arent same length")

        self.batch_size = len(input_ids)
        if not (
            [len(example) for example in input_ids]
            == [len(example) for example in attention_mask]
            == [len(example) for example in entity_ids]
        ):
            raise ValueError("Examples in batch arent same length")

        self.input_ids = input_ids
        self.attention_mask = attention_mask
        self.entity_ids = entity_ids
        self.context_id = context_id


class BIOPrediction:
    def __init__(self, context_id: int, entity_id: int, label: str, score: float):
        self.context_id = context_id
        self.entity_id = entity_id
        self.label = label
        self.score = score


class HFBIOTaggerPredictor(BasePredictor):
    """
    This is a generic wrapper around Huggingface Token Classification models.

    First, we need an `entity_name` which defines the Document field that we will
    receive a prediction. For example, if `entity_name` is 'tokens', then we expect
    to classify every Document.token.  But technically, `entity_name` could be anything,
    such as `words` or `rows` or any Entity.

    Second, we need a `context_name` which defines the Document field that we will
    treat as an "example" that we want to run our model over. For example, if
    `context_name` is 'pages', then we'll loop over each page, running our classifier
    over all the 'tokens' in each page. If the `context_name` is `bib_entries`, then we'll
    loop over each bib entry, running our classifier over the 'tokens' in each page.

    The key consequence of defining a `context_name` is, when the model constructs batches
    of sequences that fit within the Huggingface transformer's window, it will *not*
    mix sequences from different contexts into the same batch. # BEN: I don't get this.

    @kylel
    """

    REQUIRED_BACKENDS = ["transformers", "torch", "smashed"]

    _INPUT_FIELD_NAME = "inputs"
    _CONTEXT_ID = "context_id"

    _HF_RESERVED_INPUT_IDS = "input_ids"
    _HF_RESERVED_ATTN_MASK = "attention_mask"
    _HF_RESERVED_WORD_IDS = "word_ids"
    _HF_RESERVED_WORD_PAD_VALUE = -1

    def __init__(
        self,
        model: Any,
        config: Any,
        tokenizer: Any,
        entity_name: str,
        context_name: str,
        batch_size: Optional[int] = 2,
        device: Optional[str] = "cpu",
    ):
        self.model = model
        self.config = config
        self.tokenizer = tokenizer
        self.entity_name = entity_name
        self.context_name = context_name
        self.batch_size = batch_size

        # move model to device
        self.model.to(device)

        # For some reason, the roberta max_position_embeddings is larger than the actual model context window
        # so use the model_max_length instead. (We can't replace model_max_length with max_position_embeddings
        # in general, because model_max_legnth will be set to a very large number if the model length is not
        # explicitly included in the tokenizer (like in the case of alleani/scibert). As a compromise, check
        # if the model_max_legnth is big, and it's too big, then use the max_position_embeddings as the context
        # window size.
        if tokenizer.model_max_length > 1e29:
            max_length = self.config.max_position_embeddings
        else:
            max_length = tokenizer.model_max_length

        # handles tokenization, sliding window, truncation, subword to input word mapping, etc.
        self.tokenizer_mapper = TokenizerMapper(
            input_field=self._INPUT_FIELD_NAME,
            tokenizer=tokenizer,
            is_split_into_words=True,
            add_special_tokens=True,
            truncation=True,
            max_length=max_length,
            return_overflowing_tokens=True,
            return_word_ids=True,
        )
        # since input data is automatically chunked into segments (e.g. 512 length),
        # each example <dict> actually becomes many input sequences.
        # this mapper unpacks all of this into one <dict> per input sequence.
        # we set `repeat` because we want other fields (`context_id`) to repeat across sequnces
        # E.g. {"input_ids": [[1,2,3]]} => {"input_ids": [1,2,3]}
        self.unpacking_mapper = UnpackingMapper(
            fields_to_unpack=[
                self._HF_RESERVED_INPUT_IDS,
                self._HF_RESERVED_ATTN_MASK,
                self._HF_RESERVED_WORD_IDS,
            ],
            ignored_behavior="repeat",
        )
        # at the end of this, each <dict> contains <lists> of length `batch_size`
        # where each element is variable length within the `max_length` limit.
        # `keep_last` controls whether we want partial batches, which we always do
        # for token classification (i.e. we dont want to miss any token!)
        self.batch_size_mapper = FixedBatchSizeMapper(batch_size=batch_size, keep_last=True)
        # this performs padding so all sequences in a batch are of same length
        self.list_collator_mapper = FromTokenizerListCollatorMapper(
            tokenizer=tokenizer,
            pad_to_length=None,  # keeping this `None` is best because of dynamic padding
            fields_pad_ids={
                self._HF_RESERVED_WORD_IDS: self._HF_RESERVED_WORD_PAD_VALUE,
                "labels": -100,
            },
        )
        # this casts python Dict[List] into tensors.  if using GPU, would do `device='gpu'`
        self.python_to_torch_mapper = Python2TorchMapper(device=self.model.device)
        # combining everything
        self.preprocess_mapper = self.tokenizer_mapper >> self.unpacking_mapper >> self.batch_size_mapper

    @property
    def device(self) -> torch.device:
        device = self.model.device
        device = torch.device(device) if isinstance(device, str) else device
        return device

    @device.setter
    def device(self, device: Union[torch.device, str]) -> None:
        device = torch.device(device) if isinstance(device, str) else device
        self.model.to(device)
        self.python_to_torch_mapper.device = device

    @classmethod
    def from_pretrained(
        cls,
        model_name_or_path: str,
        entity_name: str,
        context_name: str,
        batch_size: Optional[int] = 2,
        device: Optional[str] = "cpu",
        *args,
        **kwargs,
    ):
        """If `model_name_or_path` is a path, should be a directory
        containing `vocab.txt`, `config.json`, and `pytorch_model.bin`

        NOTE: slightly annoying, but if loading in this way, the `_name_or_path`
        in `model.config` != `config`.
        """
        tokenizer = transformers.AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=model_name_or_path, *args, **kwargs
        )

        if "add_prefix_space" in kwargs:
            kwargs.pop("add_prefix_space")

        config = transformers.AutoConfig.from_pretrained(
            pretrained_model_name_or_path=model_name_or_path, *args, **kwargs
        )
        model = transformers.AutoModelForTokenClassification.from_pretrained(
            pretrained_model_name_or_path=model_name_or_path, *args, **kwargs
        )
        predictor = cls(
            model=model,
            config=config,
            tokenizer=tokenizer,
            entity_name=entity_name,
            context_name=context_name,
            batch_size=batch_size,
            device=device,
        )
        return predictor

    @property
    def REQUIRED_DOCUMENT_FIELDS(self) -> List[str]:
        return [self.context_name, self.entity_name]

    def preprocess(self, doc: Document, context_name: str) -> List[BIOBatch]:
        """Processes document into whatever makes sense for the Huggingface model"""
        # (1) get it into a dictionary format that Smashed expects

        dataset = [
            {
                self._INPUT_FIELD_NAME: [entity.text for entity in getattr(context, self.entity_name)],
                self._CONTEXT_ID: i,
            }
            for i, context in enumerate(getattr(doc, context_name))
        ]

        # (2) apply Smashed
        batch_dicts = self.preprocess_mapper.map(dataset=dataset)

        # (3) convert dicts to objects
        return [
            # slightly annoying, but the names `input_ids`, `attention_mask` and `word_ids` are
            # reserved and produced after tokenization, which is why hard-coded here.
            BIOBatch(
                input_ids=batch_dict[self._HF_RESERVED_INPUT_IDS],
                attention_mask=batch_dict[self._HF_RESERVED_ATTN_MASK],
                entity_ids=batch_dict[self._HF_RESERVED_WORD_IDS],
                context_id=batch_dict[self._CONTEXT_ID],
            )
            for batch_dict in batch_dicts
        ]

    def combine_annotations(self, annotations: List[Entity]) -> Annotation:
        # import pytest; pytest.set_trace()
        label = annotations[0].metadata.label[2:]  # remove the B-
        scores = [annotation.metadata.score for annotation in annotations]
        span = Span.create_enclosing_span([span for annotation in annotations for span in annotation.spans])
        smaller_boxes = [box for annotation in annotations for box in annotation.boxes]
        box = [Box.create_enclosing_box(smaller_boxes)] if smaller_boxes else []
        return Entity(
            spans=[span],
            boxes=box,
            metadata=Metadata(label=label, score=np.mean(scores)),
        )

    def postprocess(self, doc: Document, context_name: str, preds: List[BIOPrediction]) -> List[Annotation]:
        """This function handles a bunch of nonsense that happens with Huggingface models &
        how we processed the data.  Namely:

        Because Huggingface might drop tokens during the course of tokenization
        we need to organize our predictions into a Lookup <dict> and cross-reference
        with the original input SpanGroups to make sure they all got classified.
        """
        # (1) organize predictions into a Lookup at the (Context, SpanGroup) level.
        context_id_to_entity_id_to_pred: Dict[int, Dict[int, BIOPrediction]] = defaultdict(dict)
        for prediction in preds:
            context_id_to_entity_id_to_pred[prediction.context_id][prediction.entity_id] = prediction

        # (2) iterate through original data to check against that Lookup
        annotations: List[Annotation] = []
        field_annotations: List[Entity] = []
        for i, context in enumerate(getattr(doc, context_name)):
            for j, entity in enumerate(getattr(context, self.entity_name)):
                pred: Optional[BIOPrediction] = context_id_to_entity_id_to_pred[i].get(j, None)
                # TODO: double-check whether this deepcopy is needed...
                new_metadata = Metadata.from_json(entity.metadata.to_json())
                if pred is not None:
                    new_metadata.label = pred.label
                    new_metadata.score = pred.score
                else:
                    new_metadata.label = None
                    new_metadata.score = None
                new_entity = Entity(
                    spans=entity.spans,
                    boxes=entity.boxes,
                    metadata=new_metadata,
                )

                if new_entity.metadata.label.startswith("B"):
                    # end the last annotation
                    if field_annotations:
                        annotations.append(self.combine_annotations(field_annotations))
                    field_annotations = [new_entity]
                elif new_entity.metadata.label == "O":
                    # end the last annotation
                    if field_annotations:
                        annotations.append(self.combine_annotations(field_annotations))
                    field_annotations = []
                else:
                    # Either "I", so we're in the same field or "None", so we're in the same word-piece
                    field_annotations.append(new_entity)
                # annotations.append(new_entity)
        if field_annotations:
            annotations.append(self.combine_annotations(field_annotations))

        return annotations

    def _predict(self, doc: Document) -> List[Annotation]:
        # (1) Make batches
        batches: List[BIOBatch] = self.preprocess(doc=doc, context_name=self.context_name)

        # (2) Predict each batch.
        preds: List[BIOPrediction] = []
        for batch in batches:
            for pred in self._predict_batch(batch=batch):
                preds.append(pred)

        # (3) Postprocess into proper Annotations
        annotations = self.postprocess(doc=doc, context_name=self.context_name, preds=preds)
        return annotations

    def _predict_batch(
        self,
        batch: BIOBatch,
        device: Union[None, str, torch.device] = None
    ) -> List[BIOPrediction]:
        #
        #   preprocessing!!  (padding & tensorification)
        #
        pytorch_batch = self.python_to_torch_mapper.transform(
            data=self.list_collator_mapper.transform(
                data={
                    self._HF_RESERVED_INPUT_IDS: batch.input_ids,
                    self._HF_RESERVED_ATTN_MASK: batch.attention_mask,
                }
            )
        )
        # change device if needed
        if device:
            pytorch_batch = {k: v.to(device) for k, v in pytorch_batch.items()}
        #
        #   inference!! (preferably on gpu)
        #
        self.model.eval()
        with torch.no_grad():
            pytorch_output = self.model(**pytorch_batch)
        scores_tensor = torch.softmax(pytorch_output.logits, dim=2)
        token_scoresss = [
            [token_scores for token_scores, yn in zip(token_scoress, yns) if yn == 1]
            for token_scoress, yns in zip(scores_tensor.tolist(), batch.attention_mask)
        ]
        #
        #   postprocessing (map back to original inputs)!!
        #
        preds = []
        for j, (context_id, word_ids, token_scoress) in enumerate(
            zip(batch.context_id, batch.entity_ids, token_scoresss)
        ):
            for word_id, token_scores, is_valid_pred in zip(
                word_ids, token_scoress, self._token_pooling_strategy_mask(word_ids=word_ids)
            ):
                if word_id is None or is_valid_pred is False:
                    continue
                else:
                    label_id = np.argmax(token_scores)
                    pred = BIOPrediction(
                        context_id=context_id,
                        entity_id=word_id,
                        label=self.config.id2label[label_id],
                        score=token_scores[label_id],
                    )
                    preds.append(pred)
        return preds

    def _token_pooling_strategy_mask(
        self,
        token_ids: Optional[List[int]] = None,
        word_ids: Optional[List[Optional[int]]] = None,
        token_scores: Optional[List[Tuple[float, float]]] = None,
        strategy: str = "first",
    ) -> List[bool]:
        """
        words are split into multiple tokens, each of which has a prediction.
        there are multiple strategies to decide the model prediction at a word-level:
        1) 'first': take only the first token prediction for whole word
        2) 'max': take the highest scoring token prediction for whole word
        3) ...
        """
        if word_ids is None:
            word_ids = []

        if strategy == "first":
            mask = [True]
            prev_word_id = word_ids[0]
            for current_word_id in word_ids[1:]:
                if current_word_id == prev_word_id:
                    mask.append(False)
                else:
                    mask.append(True)
                prev_word_id = current_word_id
        else:
            raise NotImplementedError(f"mode {strategy} not implemented yet")

        # if no word ID (e.g. [cls], [sep]), always mask
        mask = [is_word if word_id is not None else False for is_word, word_id in zip(mask, word_ids)]
        return mask



"""


@lucas

"""

from copy import deepcopy
from dataclasses import MISSING, Field, fields, is_dataclass
from functools import wraps
import inspect
import logging

from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

__all__ = ["store_field_in_metadata", "Metadata"]


class _DEFAULT:
    # object to keep track if a default value is provided when getting
    # or popping a key from Metadata
    ...


class Metadata:
    """An object that contains metadata for an annotation.
    It supports dot access and dict-like access."""

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            self.set(k, v)

    @overload
    def get(self, key: str) -> Any:
        """Get value with name `key` in metadata"""
        ...

    @overload
    def get(self, key: str, default: Any) -> Any:
        """Get value with name `key` in metadata"""
        ...

    def get(self, key: str, default: Optional[Any] = _DEFAULT) -> Any:
        """Get value with name `key` in metadata;
        if not found, return `default` if specified,
        otherwise raise `None`"""
        if key in self.__dict__:
            return self.__dict__[key]
        elif default != _DEFAULT:
            return default
        else:
            logging.warning(f"{key} not found in metadata")
            return None

    def has(self, key: str) -> bool:
        """Check if metadata contains key `key`; return `True` if so,
        `False` otherwise"""
        return key in self.__dict__

    def set(self, key: str, value: Any) -> None:
        """Set `key` in metadata to `value`; key must be a valid Python
        identifier (that is, a valid variable name) otherwise,
        raise a ValueError"""
        if not key.isidentifier():
            raise ValueError(
                f"`{key}` is not a valid variable name, "
                "so it cannot be used as key in metadata"
            )
        self.__dict__[key] = value

    @overload
    def pop(self, key: str) -> Any:
        """Remove & returns value for `key` from metadata;
        raise `KeyError` if not found"""
        ...

    @overload
    def pop(self, key: str, default: Any) -> Any:
        """Remove & returns value for `key` from metadata;
        if not found, return `default`"""
        ...

    def pop(self, key: str, default: Optional[Any] = _DEFAULT) -> Any:
        """Remove & returns value for `key` from metadata;
        if not found, return `default` if specified,
        otherwise raise `KeyError`"""
        if key in self.__dict__:
            return self.__dict__.pop(key)
        elif default != _DEFAULT:
            return default
        else:
            raise KeyError(f"{key} not found in metadata")

    def keys(self) -> Iterable[str]:
        """Return an iterator over the keys in metadata"""
        return self.__dict__.keys()

    def values(self) -> Iterable[Any]:
        """Return an iterator over the values in metadata"""
        return self.__dict__.values()

    def items(self) -> Iterable[Tuple[str, Any]]:
        """Return an iterator over <key, value> pairs in metadata"""
        return self.__dict__.items()

    # Interfaces from loading/saving to dictionary
    def to_json(self) -> Dict[str, Any]:
        """Return a dict representation of metadata"""
        return deepcopy(self.__dict__)

    @classmethod
    def from_json(cls, di: Dict[str, Any]) -> "Metadata":
        """Create a Metadata object from a dict representation"""
        metadata = cls()
        for k, v in di.items():
            metadata.set(k, v)
        return metadata

    # The following methods are to ensure equality between metadata
    # with same keys and values
    def __len__(self) -> int:
        return len(self.__dict__)

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, Metadata):
            return False
        if len(self) != len(__o):
            return False
        for k in __o.keys():
            if k not in self.keys() or self[k] != __o[k]:
                return False
        return True

    # The following methods are for compatibility with the dict interface
    def __contains__(self, key: str) -> bool:
        return self.has(key)

    def __iter__(self) -> Iterable[str]:
        return self.keys()

    def __getitem__(self, key: str) -> Any:
        return self.get(key)

    def __setitem__(self, key: str, value: Any) -> None:
        return self.set(key, value)

    # The following methods are for compatibility with the dot access interface
    def __getattr__(self, key: str) -> Any:
        return self.get(key)

    def __setattr__(self, key: str, value: Any) -> None:
        return self.set(key, value)

    def __delattr__(self, key: str) -> Any:
        return self.pop(key)

    # The following methods return a nice representation of the metadata
    def __repr__(self) -> str:
        return f"Metadata({repr(self.__dict__)})"

    def __str__(self) -> str:
        return f"Metadata({str(self.__dict__)})"

    # Finally, we need to support pickling/copying
    def __deepcopy__(self, memo: Dict[int, Any]) -> "Metadata":
        return Metadata.from_json(deepcopy(self.__dict__, memo))


T = TypeVar("T")


def store_field_in_metadata(
    field_name: str,
    getter_fn: Optional[Callable[[T], Any]] = None,
    setter_fn: Optional[Callable[[T, Any], None]] = None,
) -> Callable[[Type[T]], Type[T]]:
    """This decorator is used to store a field that was previously part of
    a dataclass in a metadata attribute, while keeping it accessible as an
    attribute using the .field_name notation. Example:

    @store_field_in_metadata('field_a')
    @dataclass
    class MyDataclass:
        metadata: Metadata = field(default_factory=Metadata)
        field_a: int = 3
        field_b: int = 4

    d = MyDataclass()
    print(d.field_a) # 3
    print(d.metadata)   # Metadata({'field_a': 3})
    print(d) # MyDataclass(field_a=3, field_b=4, metadata={'field_a': 3})

    d = MyDataclass(field_a=5)
    print(d.field_a) # 5
    print(d.metadata)   # Metadata({'field_a': 5})
    print(d) # MyDataclass(field_a=5, field_b=4, metadata={'field_a': 5})

    Args:
        field_name: The name of the field to store in the metadata.
        getter_fn: A function that takes the dataclass instance and
            returns the value of the field. If None, the field's value is
            looked up from the metadata dictionary.
        setter_fn: A function that takes the dataclass instance and a value
            for the field and sets it. If None, the field's value is added
            to the metadata dictionary.
    """

    def wrapper_fn(
        cls_: Type[T],
        wrapper_field_name: str = field_name,
        wrapper_getter_fn: Optional[Callable[[T], Any]] = getter_fn,
        wrapper_setter_fn: Optional[Callable[[T, Any], None]] = setter_fn,
    ) -> Type[T]:
        """
        This wrapper consists of three steps:
        1. Basic checks to determine if a field can be stored in the metadata.
           This includes checking that cls_ is a dataclass, that cls_ has a
           metadata attribute, and that the field is a field of cls_.
        2. Wrap the init method of cls_ to ensure that, if a field is specified
           in the metadata, it is *NOT* overwritten by the default value of the
           field. (keep reading through this function for more details)
        3. Create getter and setter methods for the field; these are going to
           override the original attribute and will be responsible for querying
           the metadata for the value of the field.

        Args:
            cls_ (Type[T]): The dataclass to wrap.
            wrapper_field_name (str): The name of the field to store in the
                metadata.
            wrapper_getter_fn (Optional[Callable[[T], Any]]): A function that
                takes returns the value of the field. If None, the getter
                is a simple lookup in the metadata dictionary.
            wrapper_setter_fn (Optional[Callable[[T, Any], None]]): A function
                that is used to set the value of the field. If None, the setter
                is a simple addition to the metadata dictionary.
        """

        # # # # # # # # # # # # STEP 1: BASIC CHECKS # # # # # # # # # # # # #
        if not (is_dataclass(cls_)):
            raise TypeError("add_deprecated_field only works on dataclasses")

        dataclass_fields = {field.name: field for field in fields(cls_)}

        # ensures we have a metadata dict where to store the field value
        if "metadata" not in dataclass_fields:
            raise TypeError(
                "add_deprecated_field requires a `metadata` field"
                "in the dataclass of type dict."
            )

        if not issubclass(dataclass_fields["metadata"].type, Metadata):
            raise TypeError(
                "add_deprecated_field requires a `metadata` field "
                "in the dataclass of type Metadata, not "
                f"{dataclass_fields['metadata'].type}."
            )

        # ensure the field is declared in the dataclass
        if wrapper_field_name not in dataclass_fields:
            raise TypeError(
                f"add_deprecated_field requires a `{wrapper_field_name}` field"
                "in the dataclass."
            )
        # # # # # # # # # # # # # # END OF STEP 1 # # # # # # # # # # # # # # #

        # # # # # # # # # # # # # STEP 2: WRAP INIT # # # # # # # # # # # # # #
        # In the following comment, we explain the need for step 2.
        #
        # We want to make sure that if a field is specified in the metadata,
        # the default value of the field provided during class annotation does
        # not override it. For example, consider the following code:
        #
        #   @store_field_in_metadata('field_a')
        #   @dataclass
        #   class MyDataclass:
        #       metadata: Metadata = field(default_factory=Metadata)
        #       field_a: int = 3
        #
        # If we don't disable wrap the __init__ method, the following code
        # will print `3`
        #
        #   d = MyDataclass(metadata={'field_a': 5})
        #   print(d.field_a)
        #
        # but if we do, it will work fine and print `5` as expected.
        #
        # The reason why this occurs is that the __init__ method generated
        # by a dataclass uses the default value of the field to initialize the
        # class if a default is not provided.
        #
        # Our solution is rather simple: before calling the dataclass init,
        # we look if:
        #   1. A `metadata` argument is provided in the constructor, and
        #   2. The `metadata` argument contains a field with name ``
        #
        # To disable the auto-init, we have to do two things:
        #   1. create a new dataclass that inherits from the original one,
        #      but with init=False for field wrapper_field_name
        #   2. create a wrapper for the __init__ method of the new dataclass
        #      that, when called, calls the original __init__ method and then
        #      adds the field value to the metadata dict.

        # This signature is going to be used to bind to the args/kwargs during
        # init, which allows easy lookup of arguments/keywords arguments by
        # name.
        cls_signature = inspect.signature(cls_.__init__)

        # We need to save the init method since we will override it.
        cls_init_fn = cls_.__init__

        @wraps(cls_init_fn)
        def init_wrapper(self, *args, **kwargs):
            # parse the arguments and keywords arguments
            arguments = cls_signature.bind(self, *args, **kwargs).arguments

            # this adds the metadata to kwargs if it is not already there
            metadata = arguments.setdefault("metadata", Metadata())

            # this is the main check:
            # (a) the metadata argument contains the field we are storing in
            # the metadata, and (b) the field is not in args/kwargs, then we
            # pass the field value in the metadata to the original init method
            # to prevent it from being overwritten by its default value.
            if (
                wrapper_field_name in metadata
                and wrapper_field_name not in arguments
            ):
                arguments[wrapper_field_name] = metadata[wrapper_field_name]

            # type: ignore is due to pylance not recognizing that the
            # arguments in the signature contain a `self` key
            cls_init_fn(**arguments)  # type: ignore

        setattr(cls_, "__init__", init_wrapper)
        # # # # # # # # # # # # # # END OF STEP 2 # # # # # # # # # # # # # # #

        # # # # # # # # # # # STEP 3: GETTERS & SETTERS # # # # # # # # # # # #

        # We add the getter from here on:
        if wrapper_getter_fn is None:
            # create property for the deprecated field, as well as a setter
            # that will add to the underlying metadata dict
            def _wrapper_getter_fn(
                self, field_spec: Field = dataclass_fields[wrapper_field_name]
            ):
                # we expect metadata to be of type Metadata
                metadata: Union[Metadata, None] = getattr(
                    self, "metadata", None
                )
                if metadata is not None and wrapper_field_name in metadata:
                    return metadata.get(wrapper_field_name)
                elif field_spec.default is not MISSING:
                    return field_spec.default
                elif field_spec.default_factory is not MISSING:
                    return field_spec.default_factory()
                else:
                    raise AttributeError(
                        f"Value for attribute '{wrapper_field_name}' "
                        "has not been set."
                    )

            # this avoids mypy error about redefining an argument
            wrapper_getter_fn = _wrapper_getter_fn

        field_property = property(wrapper_getter_fn)

        # We add the setter from here on:
        if wrapper_setter_fn is None:

            def _wrapper_setter_fn(self: T, value: Any) -> None:
                # need to use getattr otherwise pylance complains
                # about not knowing if 'metadata' is available as an
                # attribute of self (which it is, since we checked above
                # that it is in the dataclass fields)
                metadata: Union[Metadata, None] = getattr(
                    self, "metadata", None
                )

                if metadata is None:
                    raise RuntimeError(
                        "all deprecated fields must be declared after the "
                        f"`metadata` field; however, `{wrapper_field_name}`"
                        " was declared before. Fix your class definition."
                    )

                metadata.set(wrapper_field_name, value)

            # this avoids mypy error about redefining an argument
            wrapper_setter_fn = _wrapper_setter_fn

        # make a setter for the deprecated field
        field_property = field_property.setter(wrapper_setter_fn)

        # assign the property to the dataclass
        setattr(cls_, wrapper_field_name, field_property)
        # # # # # # # # # # # # # # END OF STEP 3 # # # # # # # # # # # # # # #

        return cls_

    return wrapper_fn



import itertools
import string
from typing import List, Optional, Tuple, Union

import pdfplumber

try:
    # pdfplumber >= 0.8.0
    import pdfplumber.utils.text as ppu
except:
    # pdfplumber <= 0.7.6
    import pdfplumber.utils as ppu

from papermage.magelib import (
    Box,
    Document,
    EntitiesFieldName,
    Entity,
    Metadata,
    PagesFieldName,
    RowsFieldName,
    Span,
    SymbolsFieldName,
    TokensFieldName,
)
from papermage.parsers.parser import Parser
from papermage.utils.text import maybe_normalize

_TOL = Union[int, float]


class WordExtractorWithFontInfo(ppu.WordExtractor):
    """Override WordExtractor methods to append additional char-level info."""

    def __init__(
        self,
        x_tolerance: _TOL = ppu.DEFAULT_X_TOLERANCE,
        y_tolerance: _TOL = ppu.DEFAULT_Y_TOLERANCE,
        keep_blank_chars: bool = False,
        use_text_flow: bool = False,
        horizontal_ltr: bool = True,
        vertical_ttb: bool = True,
        extra_attrs: Optional[List[str]] = None,
        split_at_punctuation: Union[bool, str] = False,
        append_attrs: Optional[List[str]] = None,
    ):
        super().__init__(
            x_tolerance=x_tolerance,
            y_tolerance=y_tolerance,
            keep_blank_chars=keep_blank_chars,
            use_text_flow=use_text_flow,
            horizontal_ltr=horizontal_ltr,
            vertical_ttb=vertical_ttb,
            extra_attrs=extra_attrs,
            split_at_punctuation=split_at_punctuation,
        )
        self.append_attrs = append_attrs

    def merge_chars(self, ordered_chars: ppu.T_obj_list) -> ppu.T_obj:
        """Modify returned word to capture additional char-level info."""
        word = super().merge_chars(ordered_chars=ordered_chars)

        # ordered_chars is a list of characters for the word with extra attributes.
        # Here we simply append additional information to the word using the first char.
        if self.append_attrs is not None:
            for key in self.append_attrs:
                if key not in word:
                    word[key] = ordered_chars[0][key]

        return word

    def extract(self, chars: ppu.T_obj_list) -> ppu.T_obj_list:
        if hasattr(super(), "extract"):
            # pdfplumber <= 0.7.6
            return super().extract(chars)
        else:
            # pdfplumber >= 0.8.0
            return super().extract_words(chars)


class PDFPlumberParser(Parser):
    # manually added characters: '–' and '§'
    DEFAULT_PUNCTUATION_CHARS = string.punctuation + chr(8211) + chr(167)

    def __init__(
        self,
        token_x_tolerance: _TOL = 1.5,
        token_y_tolerance: _TOL = 2,
        line_x_tolerance: _TOL = 10,
        line_y_tolerance: _TOL = 10,
        keep_blank_chars: bool = False,
        use_text_flow: bool = True,
        horizontal_ltr: bool = True,
        vertical_ttb: bool = True,
        extra_attrs: Optional[List[str]] = None,
        split_at_punctuation: Union[str, bool] = True,
    ):
        """The PDFPlumber PDF Detector

        Args:
            token_x_tolerance (_TOL, optional):
                The threshold used for extracting "word tokens" from the pdf file.
                It will merge the pdf characters into a word token if the difference
                between the x_2 of one character and the x_1 of the next is less than
                or equal to token_x_tolerance. See details in `pdf2plumber's documentation
                <https://github.com/jsvine/pdfplumber#the-pdfplumberpage-class>`_.
                Defaults to 1.5, in absolute coordinates.
            token_y_tolerance (_TOL, optional):
                The threshold used for extracting "word tokens" from the pdf file.
                It will merge the pdf characters into a word token if the difference
                between the y_2 of one character and the y_1 of the next is less than
                or equal to token_y_tolerance. See details in `pdf2plumber's documentation
                <https://github.com/jsvine/pdfplumber#the-pdfplumberpage-class>`_.
                Defaults to 2, in absolute coordinates.
            line_x_tolerance (_TOL, optional):
                The threshold used for extracting "line tokens" from the pdf file.
                Defaults to 10, in absolute coordinates.
            line_y_tolerance (_TOL, optional):
                The threshold used for extracting "line tokens" from the pdf file.
                Defaults to 10, in absolute coordinates.
            keep_blank_chars (bool, optional):
                When keep_blank_chars is set to True, it will treat blank characters
                are treated as part of a word, not as a space between words. See
                details in `pdf2plumber's documentation
                <https://github.com/jsvine/pdfplumber#the-pdfplumberpage-class>`_.
                Defaults to False.
            use_text_flow (bool, optional):
                When use_text_flow is set to True, it will use the PDF's underlying
                flow of characters as a guide for ordering and segmenting the words,
                rather than presorting the characters by x/y position. (This mimics
                how dragging a cursor highlights text in a PDF; as with that, the
                order does not always appear to be logical.) See details in
                `pdf2plumber's documentation
                <https://github.com/jsvine/pdfplumber#the-pdfplumberpage-class>`_.
                Defaults to True.
            horizontal_ltr (bool, optional):
                When horizontal_ltr is set to True, it means the doc should read
                text from left to right, vice versa.
                Defaults to True.
            vertical_ttb (bool, optional):
                When vertical_ttb is set to True, it means the doc should read
                text from top to bottom, vice versa.
                Defaults to True.
            extra_attrs (Optional[List[str]], optional):
                Passing a list of extra_attrs (e.g., ["fontname", "size"]) will
                restrict each words to characters that share exactly the same
                value for each of those `attributes extracted by pdfplumber
                <https://github.com/jsvine/pdfplumber/blob/develop/README.md#char-properties>`_,
                and the resulting word dicts will indicate those attributes.
                See details in `pdf2plumber's documentation
                <https://github.com/jsvine/pdfplumber#the-pdfplumberpage-class>`_.
                Defaults to `["fontname", "size"]`.
        """
        self.token_x_tolerance = token_x_tolerance
        self.token_y_tolerance = token_y_tolerance
        self.line_x_tolerance = line_x_tolerance
        self.line_y_tolerance = line_y_tolerance
        self.keep_blank_chars = keep_blank_chars
        self.use_text_flow = use_text_flow
        self.horizontal_ltr = horizontal_ltr
        self.vertical_ttb = vertical_ttb
        self.extra_attrs = extra_attrs if extra_attrs is not None else ["fontname", "size"]
        if split_at_punctuation is True:
            split_at_punctuation = type(self).DEFAULT_PUNCTUATION_CHARS
        self.split_at_punctuation = split_at_punctuation

    def parse(self, input_pdf_path: str, **kwargs) -> Document:
        with pdfplumber.open(input_pdf_path) as plumber_pdf_object:
            all_tokens = []
            all_word_ids = []
            last_word_id = -1
            all_row_ids = []
            last_row_id = -1
            all_page_ids = []
            all_page_dims = []

            for page_id, page in enumerate(plumber_pdf_object.pages):
                page_unit = float(page.page_obj.attrs.get("UserUnit", 1.0))
                all_page_dims.append((float(page.width), float(page.height), page_unit))

                # 1) tokens we use for Document.symbols
                coarse_tokens = page.extract_words(
                    x_tolerance=self.token_x_tolerance,
                    y_tolerance=self.token_y_tolerance,
                    keep_blank_chars=self.keep_blank_chars,
                    use_text_flow=self.use_text_flow,
                    horizontal_ltr=self.horizontal_ltr,
                    vertical_ttb=self.vertical_ttb,
                    extra_attrs=self.extra_attrs,
                    split_at_punctuation=None,
                )
                # 2) tokens we use for Document.tokens
                fine_tokens = WordExtractorWithFontInfo(
                    x_tolerance=self.token_x_tolerance,
                    y_tolerance=self.token_y_tolerance,
                    keep_blank_chars=self.keep_blank_chars,
                    use_text_flow=self.use_text_flow,
                    horizontal_ltr=self.horizontal_ltr,
                    vertical_ttb=self.vertical_ttb,
                    extra_attrs=self.extra_attrs,
                    split_at_punctuation=self.split_at_punctuation,
                    append_attrs=["fontname", "size"],
                ).extract(page.chars)
                # 3) align fine tokens with coarse tokens
                word_ids_of_fine_tokens = self._align_coarse_and_fine_tokens(
                    coarse_tokens=[c["text"] for c in coarse_tokens],
                    fine_tokens=[f["text"] for f in fine_tokens],
                )
                assert len(word_ids_of_fine_tokens) == len(fine_tokens)
                # 4) normalize / clean tokens & boxes
                fine_tokens = [
                    {
                        "text": token["text"],
                        "fontname": maybe_normalize(token["fontname"]),
                        "size": token["size"],
                        "bbox": Box.from_xy_coordinates(
                            x1=float(token["x0"]),
                            y1=float(token["top"]),
                            x2=float(token["x1"]),
                            y2=float(token["bottom"]),
                            page_width=float(page.width),
                            page_height=float(page.height),
                            page=int(page_id),
                        ).to_relative(page_width=float(page.width), page_height=float(page.height)),
                    }
                    for token in fine_tokens
                ]
                # 5) group tokens into lines
                # TODO - doesnt belong in parser; should be own predictor
                line_ids_of_fine_tokens = self._simple_line_detection(
                    page_tokens=fine_tokens,
                    x_tolerance=self.line_x_tolerance / page.width,
                    y_tolerance=self.line_y_tolerance / page.height,
                )
                assert len(line_ids_of_fine_tokens) == len(fine_tokens)
                # 6) accumulate
                all_tokens.extend(fine_tokens)
                all_row_ids.extend([i + last_row_id + 1 for i in line_ids_of_fine_tokens])
                last_row_id = all_row_ids[-1]
                all_word_ids.extend([i + last_word_id + 1 for i in word_ids_of_fine_tokens])
                last_word_id = all_word_ids[-1]
                for _ in fine_tokens:
                    all_page_ids.append(page_id)
            # now turn into a beautiful document!
            doc_json = self._convert_nested_text_to_doc_json(
                token_dicts=all_tokens,
                word_ids=all_word_ids,
                row_ids=all_row_ids,
                page_ids=all_page_ids,
                dims=all_page_dims,
            )
            doc = Document.from_json(doc_json)
            return doc

    def _convert_nested_text_to_doc_json(
        self,
        token_dicts: List[dict],
        word_ids: List[int],
        row_ids: List[int],
        page_ids: List[int],
        dims: List[Tuple[float, float, float]],
    ) -> dict:
        """For a single page worth of text"""

        # 1) build tokens & symbols
        symbols = ""
        token_annos: List[Entity] = []
        start = 0
        for token_id in range(len(token_dicts) - 1):
            token_dict = token_dicts[token_id]
            current_word_id = word_ids[token_id]
            next_word_id = word_ids[token_id + 1]
            current_row_id = row_ids[token_id]
            next_row_id = row_ids[token_id + 1]

            # 1) add to symbols
            symbols += token_dict["text"]

            # 2) make Token
            end = start + len(token_dict["text"])

            # 2b) gather token metadata
            if "fontname" in token_dict and "size" in token_dict:
                token_metadata = Metadata(
                    fontname=token_dict["fontname"],
                    size=token_dict["size"],
                )
            else:
                token_metadata = None

            token = Entity(
                spans=[Span(start=start, end=end)],
                boxes=[token_dict["bbox"]],
                metadata=token_metadata,
            )
            token_annos.append(token)

            # 3) increment whitespace based on Row & Word membership. and build Rows.
            if next_row_id == current_row_id:
                if next_word_id == current_word_id:
                    start = end
                else:
                    symbols += " "
                    start = end + 1
            else:
                # new row
                symbols += "\n"
                start = end + 1
        # handle last token
        symbols += token_dicts[-1]["text"]
        end = start + len(token_dicts[-1]["text"])
        token = Entity(
            spans=[Span(start=start, end=end)],
            boxes=[token_dicts[-1]["bbox"]],
        )
        token_annos.append(token)

        # 2) build rows
        tokens_with_group_ids = [
            (token, row_id, page_id) for token, row_id, page_id in zip(token_annos, row_ids, page_ids)
        ]
        row_annos: List[Entity] = []
        for row_id, tups in itertools.groupby(iterable=tokens_with_group_ids, key=lambda tup: tup[1]):
            row_tokens = [token for token, _, _ in tups]
            row = Entity(
                spans=[
                    Span(
                        start=row_tokens[0].spans[0].start,
                        end=row_tokens[-1].spans[0].end,
                    )
                ],
                boxes=[Box.create_enclosing_box(boxes=[box for t in row_tokens for box in t.boxes])],
            )
            row_annos.append(row)

        # 3) build pages
        page_annos: List[Entity] = []
        for page_id, tups in itertools.groupby(iterable=tokens_with_group_ids, key=lambda tup: tup[2]):
            page_tokens = [token for token, _, _ in tups]
            page_w, page_h, page_unit = dims[page_id]
            page = Entity(
                spans=[
                    Span(
                        start=page_tokens[0].spans[0].start,
                        end=page_tokens[-1].spans[0].end,
                    )
                ],
                boxes=[Box.create_enclosing_box(boxes=[box for t in page_tokens for box in t.boxes])],
                metadata=Metadata(width=page_w, height=page_h, user_unit=page_unit),
            )
            page_annos.append(page)

        return {
            SymbolsFieldName: symbols,
            EntitiesFieldName: {
                TokensFieldName: [token.to_json() for token in token_annos],
                RowsFieldName: [row.to_json() for row in row_annos],
                PagesFieldName: [page.to_json() for page in page_annos],
            },
        }

    def _simple_line_detection(
        self, page_tokens: List[dict], x_tolerance: int = 10, y_tolerance: int = 10
    ) -> List[int]:
        """Get text lines from the page_tokens.
        It will automatically add new lines for 1) line breaks (i.e., the current token
        has a larger y_difference between the previous one than the y_tolerance) or
        2) big horizontal gaps (i.e., the current token has a larger y_difference between
        the previous one than the x_tolerance)

        Adapted from https://github.com/allenai/VILA/blob/e6d16afbd1832f44a430074855fbb4c3d3604f4a/src/vila/pdftools/pdfplumber_extractor.py#L24

        Modified Oct 2022 (kylel): Changed return value to be List[int]
        """
        prev_y = None
        prev_x = None

        lines = []
        cur_line_id = 0
        n = 0

        for token in page_tokens:
            cur_y = token["bbox"].center[1]
            cur_x = token["bbox"].xy_coordinates[0]

            if prev_y is None:
                prev_y = cur_y
                prev_x = cur_x

            if abs(cur_y - prev_y) <= y_tolerance and cur_x - prev_x <= x_tolerance:
                lines.append(cur_line_id)
                if n == 0:
                    prev_y = cur_y
                else:
                    prev_y = (prev_y * n + cur_y) / (n + 1)  # EMA of the y_height
                n += 1

            else:
                cur_line_id += 1

                lines.append(cur_line_id)
                n = 1
                prev_y = cur_y

            prev_x = token["bbox"].xy_coordinates[2]

        return lines

    def _align_coarse_and_fine_tokens(self, coarse_tokens: List[str], fine_tokens: List[str]) -> List[int]:
        """Returns a list of length len(fine_tokens) where elements of the list are
        integer indices into coarse_tokens elements."""
        assert len(coarse_tokens) <= len(fine_tokens), f"This method requires |coarse| <= |fine|"
        assert "".join(coarse_tokens) == "".join(
            fine_tokens
        ), f"This method requires the chars(coarse) == chars(fine)"

        coarse_start_ends = []
        start = 0
        for token in coarse_tokens:
            end = start + len(token)
            coarse_start_ends.append((start, end))
            start = end

        fine_start_ends = []
        start = 0
        for token in fine_tokens:
            end = start + len(token)
            fine_start_ends.append((start, end))
            start = end

        fine_id = 0
        coarse_id = 0
        out = []
        while fine_id < len(fine_start_ends) and coarse_id < len(coarse_start_ends):
            fine_start, fine_end = fine_start_ends[fine_id]
            coarse_start, coarse_end = coarse_start_ends[coarse_id]
            if coarse_start <= fine_start and fine_end <= coarse_end:
                out.append(coarse_id)
                fine_id += 1
            else:
                coarse_id += 1

        return out



"""

An annotated "unit" on a Document.

"""

from typing import Dict, List, Optional, Union

from .annotation import Annotation
from .box import Box
from .image import Image
from .metadata import Metadata
from .span import Span


class Entity(Annotation):
    def __init__(
        self,
        spans: Optional[List[Span]] = None,
        boxes: Optional[List[Box]] = None,
        images: Optional[List[Image]] = None,
        metadata: Optional[Metadata] = None,
    ):
        if not spans and not boxes:
            raise ValueError(f"At least one of `spans` or `boxes` must be set.")
        self.spans = spans if spans else []
        self.boxes = boxes if boxes else []
        self.images = images if images else []
        self.metadata = metadata if metadata else Metadata()
        super().__init__()

    def __repr__(self):
        if self.doc:
            return f"Annotated Entity:\tSpans: {True if self.spans else False}\tBoxes: {True if self.boxes else False}\nText: {self.text}"
        return f"Unannotated Entity: {self.to_json()}"

    def to_json(self) -> Dict:
        entity_dict = dict(
            spans=[span.to_json() for span in self.spans],
            boxes=[box.to_json() for box in self.boxes],
            metadata=self.metadata.to_json(),
        )
        # only serialize non-null/non-empty values
        return {k: v for k, v in entity_dict.items() if v}

    @classmethod
    def from_json(cls, entity_json: Dict) -> "Entity":
        return cls(
            spans=[Span.from_json(span_json=span_json) for span_json in entity_json.get("spans", [])],
            boxes=[Box.from_json(box_json=box_json) for box_json in entity_json.get("boxes", [])],
            metadata=Metadata.from_json(entity_json.get("metadata", {})),
        )

    @property
    def start(self) -> Union[int, float]:
        return min([span.start for span in self.spans]) if len(self.spans) > 0 else float("-inf")

    @property
    def end(self) -> Union[int, float]:
        return max([span.end for span in self.spans]) if len(self.spans) > 0 else float("inf")

    @property
    def symbols_from_spans(self) -> List[str]:
        if self.doc is not None:
            return [self.doc.symbols[span.start : span.end] for span in self.spans]
        else:
            return []

    @property
    def symbols_from_boxes(self) -> List[str]:
        if self.doc is not None:
            matched_tokens = self.doc.find_by_box(query=self, field_name="tokens")
            return [self.doc.symbols[span.start : span.end] for t in matched_tokens for span in t.spans]
        else:
            return []

    @property
    def text(self) -> str:
        # return stored metadata
        maybe_text = self.metadata.get("text", None)
        if maybe_text:
            return maybe_text
        # return derived from symbols
        if self.symbols_from_spans:
            return " ".join(self.symbols_from_spans)
        # return derived from boxes and tokens
        if self.symbols_from_boxes:
            return " ".join(self.symbols_from_boxes)
        return ""

    @text.setter
    def text(self, text: Union[str, None]) -> None:
        self.metadata.text = text

    def __iter__(self):
        """By default, iterate over the spans"""
        yield from self.spans

    def __lt__(self, other: "Entity"):
        if self.id and other.id:
            return self.id < other.id
        else:
            return self.start < other.start



import re
from typing import Dict, List, Tuple, Union, Optional
from papermage.magelib import Entity, Annotation, Prediction


def group_by(
        entities: List[Union[Entity, Annotation]],
        metadata_field: str,
        metadata_values_map: Optional[Dict[str, str]] = None
) -> Tuple[Prediction, ...]:
    """Group entities by the value of a field in metadata. After grouping,
    entities are annotated onto the document object.

    Args:
        doc (Document): the document object
        entities (List[Union[Entity, Annotation]]): the entities to group
        metadata_field (str): the field name to group by
        metadata_values (Optional[List[str]], optional): a map of names to use for the groups.
            If not provided, the metadata values will be used as the group names. Defaults to None.

    Returns:
        Document: the document object with the grouped entities
    """
    metadata_values_map = metadata_values_map or {}

    groups: Dict[str, List[Entity]] = {k: [] for k in metadata_values_map.values()}
    for entity in entities:
        mt_val = entity.metadata[metadata_field]    # type: ignore
        mt_val = metadata_values_map.get(mt_val, mt_val)

        # check if the a valid name for a layer
        if not isinstance(mt_val, str):
            raise ValueError(f'Expected metadata value to be a string, but got {type(mt_val)}: {mt_val}')
        elif re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$', mt_val) is None:
            raise ValueError(f'Expected metadata value to be a valid Python identifier, but got: {mt_val}')

        new_entity = Entity(
            boxes=getattr(entity, 'boxes', None),
            spans=getattr(entity, 'spans', None),
            metadata=getattr(entity, 'metadata', None)
        )
        groups.setdefault(mt_val, []).append(new_entity)

    return tuple(
        Prediction(name=group_name, entities=group_entities)
        for group_name, group_entities in groups.items()
    )



"""


@kylel

"""


import unittest

from papermage.magelib import Span
from papermage.magelib.span import MergeClusterSpans


class TestSpan(unittest.TestCase):
    def test_to_from_json(self):
        span = Span(start=0, end=0)
        self.assertEqual(span.to_json(), [0, 0])

        span2 = Span.from_json(span.to_json())
        self.assertEqual(span2.start, 0)
        self.assertEqual(span2.end, 0)
        self.assertListEqual(span2.to_json(), [0, 0])

    def test_sort_spans(self):
        self.assertFalse(Span(0, 0) < Span(0, 0))
        self.assertTrue(Span(0, 0) < Span(0, 1))
        self.assertTrue(Span(0, 0) < Span(1, 0))
        self.assertTrue(Span(0, 0) < Span(1, 1))
        a = Span(1, 2)
        b = Span(2, 3)
        c = Span(3, 4)
        self.assertListEqual(sorted([c, b, a]), [a, b, c])

    def test_overlap(self):
        self.assertTrue(Span(0, 0).is_overlap(other=Span(0, 0)))
        self.assertTrue(Span(0, 0).is_overlap(other=Span(0, 1)))
        self.assertTrue(Span(0, 1).is_overlap(other=Span(0, 0)))
        self.assertTrue(Span(0, 3).is_overlap(other=Span(1, 2)))
        self.assertTrue(Span(1, 2).is_overlap(other=Span(0, 3)))
        self.assertFalse(Span(0, 1).is_overlap(other=Span(1, 2)))
        self.assertFalse(Span(0, 1).is_overlap(other=Span(2, 2)))

    def test_equiv(self):
        self.assertTrue(Span(0, 0) == Span(0, 0))
        self.assertFalse(Span(0, 0) == Span(0, 1))
        self.assertFalse(Span(1, 0) == Span(0, 0))

    def test_create_enclosing_span(self):
        # nonempty
        with self.assertRaises(ValueError):
            Span.create_enclosing_span(spans=[])
        # singleton
        span = Span.create_enclosing_span(spans=[Span(1, 2)])
        self.assertEqual(span.start, 1)
        self.assertEqual(span.end, 2)
        # proper behavior
        span = Span.create_enclosing_span(spans=[Span(1, 2), Span(2, 3), Span(9, 10)])
        self.assertEqual(span.start, 1)
        self.assertEqual(span.end, 10)


class TestMergeClusterSpans(unittest.TestCase):
    def setUp(self):
        self.a = Span(0, 1)
        self.b = Span(1, 2)
        self.c = Span(3, 4)
        self.d = Span(4, 5)
        self.e = Span(9, 10)

    def test_empty(self):
        mcs = MergeClusterSpans(spans=[])
        self.assertListEqual(mcs.clusters, [[]])
        self.assertListEqual(mcs.merge(), [])

    def test_single(self):
        mcs = MergeClusterSpans(spans=[self.a])
        self.assertListEqual(mcs.clusters, [[self.a]])
        self.assertListEqual(mcs.merge(), [self.a])

    def test_disjoint(self):
        mcs = MergeClusterSpans(spans=[self.a, self.c, self.e])
        self.assertListEqual(mcs.clusters, [[self.a], [self.c], [self.e]])
        self.assertListEqual(mcs.merge(), [self.a, self.c, self.e])
        # reversed order
        mcs = MergeClusterSpans(spans=[self.e, self.c, self.a])
        self.assertListEqual(mcs.clusters, [[self.e], [self.c], [self.a]])
        self.assertListEqual(mcs.merge(), [self.e, self.c, self.a])

    def test_cluster(self):
        mcs = MergeClusterSpans(spans=[self.a, self.b, self.c, self.d, self.e], index_distance=0)
        self.assertListEqual(mcs.clusters, [[self.a, self.b], [self.c, self.d], [self.e]])
        mcs = MergeClusterSpans(spans=[self.a, self.b, self.c, self.d, self.e], index_distance=1)
        self.assertListEqual(mcs.clusters, [[self.a, self.b, self.c, self.d], [self.e]])
        mcs = MergeClusterSpans(spans=[self.a, self.b, self.c, self.d, self.e], index_distance=4)
        self.assertListEqual(mcs.clusters, [[self.a, self.b, self.c, self.d, self.e]])

    def test_merge(self):
        mcs = MergeClusterSpans(spans=[self.a, self.b, self.c, self.d, self.e], index_distance=0)
        self.assertListEqual(mcs.merge(), [Span(0, 2), Span(3, 5), Span(9, 10)])
        mcs = MergeClusterSpans(spans=[self.a, self.b, self.c, self.d, self.e], index_distance=1)
        self.assertListEqual(mcs.merge(), [Span(0, 5), Span(9, 10)])
        mcs = MergeClusterSpans(spans=[self.a, self.b, self.c, self.d, self.e], index_distance=4)
        self.assertListEqual(mcs.merge(), [Span(0, 10)])



"""

Converts pages of a PDF into Images that can be attached to a Document. 

@shannons, @kylel

"""

from typing import Iterable, List, Protocol

from papermage.magelib import Document, Image, PagesFieldName

try:
    import pdf2image
except ImportError:
    pass


class Rasterizer(Protocol):
    def rasterize(self, input_pdf_path: str, dpi: int, **kwargs) -> Iterable[Image]:
        """Given an input PDF return a List[Image]

        Args:
            input_pdf_path (str): Path to the input PDF to process
            dpi (int): Used for specify the resolution (or `DPI, dots per inch
                       <https://en.wikipedia.org/wiki/Dots_per_inch>`_) when loading images of
                       the pdf. Higher DPI values mean clearer images (also larger file sizes).

        Returns:
            Iterable[Image]
        """
        raise NotImplementedError

    @classmethod
    def attach_images(cls, images: List[Image], doc: Document) -> None:
        """Assumes doc already has `pages` annotated."""
        if PagesFieldName not in doc.fields:
            raise ValueError(f"Failed to attach. Document is missing `pages`.")
        pages = doc.get_entity(field_name=PagesFieldName)
        if len(images) != len(pages):
            raise ValueError(f"Failed to attach. {len(images)} `images` != {len(pages)} pages in `doc`.")
        for page, image in zip(pages, images):
            if page.images:
                raise AttributeError(f"Failed to attach. `images` already exists on this page {page}")
            page.images = [image]


class PDF2ImageRasterizer(Rasterizer):
    def rasterize(self, input_pdf_path: str, dpi: int, **kwargs) -> Iterable[Image]:
        """Rasterize the pdf and convert the PIL images to papermage Image objects"""
        pil_images = pdf2image.convert_from_path(pdf_path=input_pdf_path, dpi=dpi)
        images: List[Image] = []
        for pil_image in pil_images:
            image = Image()
            image.pilimage = pil_image
            images.append(image)
        return images



"""
benjaminn
"""
import dataclasses
import inspect
import json
import os
from pathlib import Path
from typing import List, Literal, Optional, Union

import pytorch_lightning as pl
import seqeval.metrics
import sklearn.metrics
import springs
import torch
import transformers
from omegaconf import DictConfig
from pytorch_lightning.callbacks import LearningRateMonitor, ModelCheckpoint
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch.utils.data import DataLoader
from tqdm import tqdm

from papermage.magelib import Box, Document, Entity, Span
from papermage.predictors import HFBIOTaggerPredictor


class HFBioTaggerPredictorWrapper(pl.LightningModule):
    def __init__(self, predictor: HFBIOTaggerPredictor):
        super().__init__()
        self.predictor = predictor
        self.learning_rate = None  # this will be set by the trainer
        self.warmup_steps = None  # this will be set by the trainer
        self.num_training_steps = None  # this will be set by the trainer

    def training_step(self, batch, batch_idx) -> Optional[STEP_OUTPUT]:
        pytorch_output = self.predictor.model(**batch)
        scores_tensor = torch.softmax(pytorch_output.logits, dim=2)
        self.log("train_loss", pytorch_output.loss.cpu(), on_step=True)
        return pytorch_output.loss

    def validation_step(self, batch, batch_idx) -> Optional[STEP_OUTPUT]:
        pytorch_output = self.predictor.model(**batch)
        self.log("val_loss", pytorch_output.loss.cpu())
        return pytorch_output.loss

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.predictor.model.parameters(), lr=self.learning_rate)
        scheduler = transformers.get_scheduler(
            "linear",
            optimizer=optimizer,
            num_warmup_steps=self.warmup_steps * self.num_training_steps,
            num_training_steps=self.num_training_steps,
        )
        return {"optimizer": optimizer, "lr_scheduler": scheduler}

    def preprocess(self, doc: Document, context_name: str = "pages") -> List:
        # is this the default context we should use?
        return self.predictor.preprocess(doc, context_name)

    def on_train_start(self) -> None:
        self.predictor.device = self.device
        return super().on_train_start()

    def on_validation_start(self) -> None:
        self.predictor.device = self.device
        return super().on_validation_start()

    def on_test_start(self) -> None:
        self.predictor.device = self.device
        return super().on_test_start()

    def __getattr__(self, name):
        """Allow access to the predictor's attributes if the wrapper doesn't have them."""
        if name not in self.__dict__:
            return getattr(self.predictor, name)


class HFCheckpoint(pl.Callback):
    def __init__(self, save_dir: str) -> None:
        """Initialize the HFCheckpoint Callback for saving huggingface checkpoints

        Args:
            save_dir (str): the directory to save the huggingface checkpoints.
        """

        super().__init__()
        self.save_dir = save_dir

    def on_save_checkpoint(self, trainer, pl_module, checkpoint):
        # save the model in hf format for easy loading
        pl_module.predictor.model.save_pretrained(self.save_dir)
        pl_module.predictor.tokenizer.save_pretrained(self.save_dir)

        # don't want to save the state twice bc it might be big. But not saving it might cause issues if we want
        # to load the model later, that's a TODO for now.
        # del checkpoint["state_dict"]


class HFBIOTaggerPredictorTrainer:
    CACHE_PATH = Path(os.environ.get("PAPERMAGE_CACHE_DIR", Path.home() / ".cache/papermage"))

    def __init__(self, predictor: HFBIOTaggerPredictor, config: DictConfig):
        if not self.CACHE_PATH.exists():
            self.CACHE_PATH.mkdir(parents=True)

        transformers.set_seed(config.seed)
        self.predictor = HFBioTaggerPredictorWrapper(predictor)
        self.predictor.learning_rate = config.learning_rate
        self.predictor.warmup_steps = config.warmup_steps

        self.config = config
        if self.config.accelerator == "auto":
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self.device = self.config.accelerator

        if config.mode == "eval":
            # just extract the original model name from the path
            # assumes that the model path starts with self.CACHE_PATH
            model_name = self.config.model_name_or_path[len(str(self.CACHE_PATH)) :].strip("/").split("_")[0]
        else:
            model_name = str(self.predictor.config.name_or_path).replace("/", "-")

        self.data_id = "_".join(
            [
                model_name,
                str(self.config.label_field),
                str(self.predictor.entity_name),
                str(self.predictor.context_name),
                str(self.predictor.batch_size),
            ]
        )

        if self.config.data_notes:
            self.data_id += f"_{self.config.data_notes}"
        (self.CACHE_PATH / self.data_id).mkdir(exist_ok=True)

        self.model_id = "_".join(
            [
                self.data_id,
                str(self.config.learning_rate),
                str(self.config.warmup_steps),
                str(self.config.seed),
                str(self.config.max_epochs),
            ]
        )

        if self.config.notes:
            self.model_id += "_" + self.config.notes

        if self.config.default_root_dir is None:
            self.config.default_root_dir = self.CACHE_PATH / self.model_id
        self.config.default_root_dir.mkdir(parents=True, exist_ok=True)

        callbacks = [HFCheckpoint(f"{self.config.default_root_dir}/checkpoints")]

        loggers = []
        if config.wandb:
            loggers.append(
                pl.loggers.WandbLogger(
                    name=self.model_id,
                    project="papermage",
                    save_dir=self.config.default_root_dir,
                )
            )

        print("Checkpoints will be saved at:", self.config.default_root_dir)

        # Filter out the args that aren't valid for pl.Trainer using the signature of __init__ function
        # This is kind of ugly...
        pl_trainer_params = inspect.signature(pl.Trainer).parameters
        pl_trainer_config = {k: v for k, v in dict(self.config).items() if k in pl_trainer_params}

        self.trainer = pl.Trainer(callbacks=callbacks, logger=loggers, **pl_trainer_config)

    def compute_labels_from_entities(self, doc: Document, labels_fields: List[str], return_labels_by_entity=False):
        """
        Computes the gold labels given a document and which field of the doc contains the label.

        if `return_labels_by_entity` is True, the labels returned will be in the same shape as the entity list:
            ie. they will be a list of labels, with one label per entity. For example, if the entity_name is
            "token", then each element of the list would be the label associated with that token.
        if `return_labels_by_entity` is False, the labels are returned in the same shape as the batches. These
            are ready to directly be fed into the Huggingface model as the labels (eg the correct tokens will
            be masked out by setting the label to -100.
        """
        batches = self.predictor.preprocess(doc, context_name=self.config.context_name)
        label_id_batches: List[List[List[int]]] = []

        # assume the label value is a binary value. (either 1, the token is in the span specified by
        # the labels_field or 0, it is not)

        # label_entities = getattr(doc, labels_field)

        # masking stategy is "first"
        # based on here: https://huggingface.co/docs/transformers/tasks/token_classification
        labels_by_entity = []
        for batch in batches:
            label_id_batch: List[List[int]] = []
            previous_entity_idx = None
            for context_idx, entity_idxs in zip(batch.context_id, batch.entity_ids):
                label_ids: List[int] = []
                for entity_idx in entity_idxs:
                    if entity_idx is None:
                        label_ids.append(-100)
                    elif entity_idx != previous_entity_idx:
                        # if the token is within the field, label it as 1
                        context = getattr(doc, self.predictor.context_name)[context_idx]
                        entity = getattr(context, self.predictor.entity_name)[entity_idx]
                        for labels_field in labels_fields:
                            overlap = getattr(entity, labels_field)
                            if overlap:
                                # assumption: overlap has at most one element (each entity has only one label associated with it)
                                # assumption: all spans are contiguous
                                if getattr(overlap[0], self.predictor.entity_name)[0] == entity:
                                    label = self.predictor.config.label2id[f"B-{labels_field}"]
                                    # label = 1  # B tag
                                else:
                                    label = self.predictor.config.label2id[f"I-{labels_field}"]
                                    # label = 2  # I tag
                                break
                        else:  # O tag
                            label = 0
                        label_ids.append(label)
                        labels_by_entity.append(label)
                    else:
                        label_ids.append(-100)
                    previous_entity_idx = entity_idx
                label_id_batch.append(label_ids)
            label_id_batches.append(label_id_batch)

        if return_labels_by_entity:
            return batches, labels_by_entity
        else:
            return batches, label_id_batches

    def preprocess(
        self, docs_path: Path, labels_fields: Optional[List[str]], annotations_path: Optional[Path] = None
    ):
        with open(docs_path, "r") as docs_file:
            all_pytorch_batches = []

            docs = [Document.from_json(json.loads(line)) for line in docs_file]
            # breakpoint()
            for doc in tqdm(docs, desc="preprocessing"):
                if labels_fields is not None:
                    batches, label_id_batches = self.compute_labels_from_entities(doc, labels_fields)

                    pytorch_batches = [
                        self.predictor.python_to_torch_mapper.transform(
                            data=self.predictor.list_collator_mapper.transform(
                                data={
                                    self.predictor._HF_RESERVED_INPUT_IDS: batch.input_ids,
                                    self.predictor._HF_RESERVED_ATTN_MASK: batch.attention_mask,
                                    "labels": label_ids,
                                }
                            )
                        )
                        for batch, label_ids in zip(batches, label_id_batches)
                    ]

                    all_pytorch_batches.extend(pytorch_batches)

            return all_pytorch_batches

    def train(
        self,
        docs_path: Path,
        annotations_entity_names: Optional[List[str]] = None,
        annotations_path: Optional[Path] = None,
    ):
        # If pytorch tensors haven't been created and cached yet
        # preprocess the document to convert it to tensors and cache them
        preprocessed_batches = []
        # cache_file = self.config.default_root_dir / "inputs.pt"
        cache_file = self.CACHE_PATH / self.data_id / "inputs.pt"
        if not cache_file.exists():
            preprocessed_batches = self.preprocess(
                docs_path=docs_path, labels_fields=annotations_entity_names, annotations_path=annotations_path
            )

            print(f"Caching preprocessed batches in: {cache_file}")
            torch.save(preprocessed_batches, cache_file)

        else:
            # load the tensors from the cache
            print(f"Loading precomputed input batches from: {cache_file}")
            preprocessed_batches = torch.load(cache_file)

        # create the dataloader
        preprocessed_batches = [{k: v.to(self.device) for k, v in batch.items()} for batch in preprocessed_batches]
        docs_dataloader = DataLoader(preprocessed_batches, batch_size=None)  # disable automatic batching

        # Run the training loop
        self._train(docs_dataloader)

    def _train(self, docs: DataLoader):
        # The module should automatically be put on the correct device, so I'm not sure why I need to do this.
        # set the number of training steps for the optimizer.
        self.predictor.num_training_steps = len(docs) * self.config.max_epochs
        self.predictor.model.to(self.device)
        self.trainer.fit(self.predictor, docs)

    def eval(self, docs_path: Path, annotations_entity_names: List[str]):
        """This is going to be a bit different from just calling `predict` on the predictor.

        This is because we might want to cache the inputs to pytorch and because we need
        to compute the labels.
        """
        self.predictor.model.to(self.device)
        docs = []
        with open(docs_path) as f:
            for line in f:
                docs.append(Document.from_json(json.loads(line)))

        all_batches = []
        all_gold_labels = []
        all_pred_labels = []
        # cache_file = self.config.default_root_dir / "test_inputs.pt"
        cache_file = self.CACHE_PATH / self.data_id / "test_inputs.pt"
        if not cache_file.exists():
            for document in tqdm(docs, desc="preprocessing", total=len(docs)):
                # Wraps the preprocess step - this should be cached
                batches, gold_labels = self.compute_labels_from_entities(
                    doc=document,
                    labels_fields=annotations_entity_names,
                    return_labels_by_entity=True,
                )
                gold_labels = [self.predictor.config.id2label[lab] for lab in gold_labels]
                all_gold_labels.append(gold_labels)
                all_batches.append(batches)
            torch.save({"batches": all_batches, "labels": all_gold_labels}, cache_file)
        else:
            cached_data = torch.load(cache_file)
            all_batches = cached_data["batches"]
            all_gold_labels = cached_data["labels"]

        all_annotations = []
        all_extracted_text = []
        for document, batches in tqdm(zip(docs, all_batches), desc="evaluating", total=len(all_batches)):
            # What I want: two lists. one of predicted labels and one of gold labels
            # each item of the list is associated with a entity (eg token)
            # basically, this amounts to removing the masked tokens from `label_ids`
            # and combining the ones with the same entity id
            # breakpoint()
            preds = []
            for batch in batches:
                for pred in self.predictor.predictor._predict_batch(batch=batch, device=self.device):
                    preds.append(pred)
            all_pred_labels.append([pred.label for pred in preds])

            # (3) Postprocess into proper Annotations
            annotations = self.predictor.predictor.postprocess(
                doc=document, context_name=self.predictor.context_name, preds=preds
            )

            all_annotations.append(annotations)

            extracted_text = []
            for annotation in annotations:
                if annotation.metadata["label"] != "O":
                    extracted_text.append(document.symbols[annotation.spans[0].start : annotation.spans[0].end])
            all_extracted_text.append(extracted_text)

        # Calculate precision, recall, f1, accuracy at the span level and token level
        try:
            clf_report_span = seqeval.metrics.classification_report(all_gold_labels, all_pred_labels)
        except ValueError:
            clf_report_span = "No predicted fields."

        # Calculate the p, r, f1, at the token level (like in the VILA paper).
        # To do this, remove the B-/I- tag from the label and just use the field name
        # (E.g. "B-Title"[:2] => "Title")
        clf_report_token = sklearn.metrics.classification_report(
            [gold_lab[2:] for gold_labels in all_gold_labels for gold_lab in gold_labels],
            [pred_lab[2:] for pred_labels in all_pred_labels for pred_lab in pred_labels],
        )

        (self.config.default_root_dir / "results").mkdir(exist_ok=True)

        with open(self.config.default_root_dir / "results" / "clf_report_token.txt", "w") as f:
            f.write(clf_report_token)

        with open(self.config.default_root_dir / "results" / "clf_report_span.txt", "w") as f:
            f.write(clf_report_span)

        # breakpoint()
        with open(self.config.default_root_dir / "results" / "results.json", "w") as f:
            json.dump(
                {
                    "y_gold": all_gold_labels,
                    "y_hat": all_pred_labels,
                    "annotations": [
                        [annotation.to_json() for annotation in annotations] for annotations in all_annotations
                    ],
                },
                f,
            )

        print("Token level prediction")
        print(clf_report_token)
        print("Results saved to:", (self.config.default_root_dir / "results"))

        # return all_gold_labels, all_pred_labels, all_extracted_text


@springs.dataclass
class HFBIOTaggerPredictorTrainConfig:
    """Stores the default training args"""

    data_path: Path  # Path to the data to train on. Should be a jsonl file where each row is a doc.
    label_field: str  # The field in the document to use as the labels for the tokens.
    # label_fields: List[str]  # The fields in the document to use as labels for the tokens.
    entity_name: str = "tokens"  # The field in the document to use as the unit of prediction.
    context_name: str = "pages"  # The field in the document to use as the input example to the model.
    model_name_or_path: str = "allenai/scibert_scivocab_uncased"
    learning_rate: float = 5e-4
    warmup_steps: float = 0.0  # proportion of training steps allocated to warmup
    mode: str = "train"  # One of "train", "eval"
    notes: str = ""
    data_notes: str = ""
    seed: int = 470

    wandb: bool = False
    # pytorch-lightning trainer args (these are the defaults). Some of the types are wrong because OmeagConf can't
    # load in arbitrary types.
    accelerator: str = "auto"
    accumulate_grad_batches: int = 1
    precision: Union[int, str] = 32
    devices: Optional[Union[str, int]] = torch.cuda.device_count() if torch.cuda.is_available() else 1
    max_epochs: int = 5
    max_steps: int = -1
    check_val_every_n_epoch: int = 1
    default_root_dir: Optional[Union[str, Path]] = None
    log_every_n_steps: int = 1
    val_check_interval: Union[float, int] = 1.0
    # Args for development/testing
    fast_dev_run: bool = False
    overfit_batches: bool = False


@springs.cli(HFBIOTaggerPredictorTrainConfig)
def main(config: HFBIOTaggerPredictorTrainConfig):
    """Launch a training run.

    For now, the simplest way to do this is to just pass the data. We can abstract this later and add some way to
    configure the training run.
    """

    # parse out the label fields:
    label_fields = config.label_field.split(",")
    labels = ["O"] + [f"{initial}-{label}" for initial in "BI" for label in label_fields]
    id2label = dict(zip(range(len(labels)), labels))
    label2id = {v: k for k, v in id2label.items()}
    print(id2label)
    print(label2id)
    kwargs = {"num_labels": len(id2label), "id2label": id2label, "label2id": label2id}

    if "roberta" in config.model_name_or_path:
        kwargs["add_prefix_space"] = True
        # kwargs["max_position_embeddings"] = 512  # This is set to 514 for some reason...

    # Initialize the trainer
    trainer = HFBIOTaggerPredictorTrainer(
        predictor=HFBIOTaggerPredictor.from_pretrained(
            model_name_or_path=config.model_name_or_path,
            entity_name=config.entity_name,
            context_name=config.context_name,
            **kwargs,
        ),
        config=config,
    )

    if config.mode == "train":
        trainer.train(docs_path=config.data_path, annotations_entity_names=label_fields)
    elif config.mode == "eval":
        trainer.eval(docs_path=config.data_path, annotations_entity_names=label_fields)
    else:
        raise ValueError(f"Invalid mode: {config.mode}")


if __name__ == "__main__":
    main()



"""

@kylel

"""

import logging
from pathlib import Path
from typing import Dict, List, Union
import warnings
from papermage.predictors.sklearn_predictors.word_predictor import make_text

from papermage.utils.annotate import group_by

from papermage.magelib import (
    AbstractsFieldName,
    AlgorithmsFieldName,
    AuthorsFieldName,
    BibliographiesFieldName,
    BlocksFieldName,
    CaptionsFieldName,
    Document,
    EntitiesFieldName,
    Entity,
    Box,
    EquationsFieldName,
    FiguresFieldName,
    FootersFieldName,
    FootnotesFieldName,
    HeadersFieldName,
    ImagesFieldName,
    KeywordsFieldName,
    ListsFieldName,
    PagesFieldName,
    ParagraphsFieldName,
    RelationsFieldName,
    RowsFieldName,
    SectionsFieldName,
    SentencesFieldName,
    SymbolsFieldName,
    TablesFieldName,
    TitlesFieldName,
    TokensFieldName,
    WordsFieldName,
)
from papermage.parsers.pdfplumber_parser import PDFPlumberParser
from papermage.predictors import (
    HFBIOTaggerPredictor,
    IVILATokenClassificationPredictor,
    LPBlockPredictor,
    PysbdSentencePredictor,
    SVMWordPredictor,
)
from papermage.rasterizers.rasterizer import PDF2ImageRasterizer
from papermage.recipes.recipe import Recipe


VILA_LABELS_MAP = {
    "Title": TitlesFieldName,
    "Paragraph": ParagraphsFieldName,
    "Author": AuthorsFieldName,
    "Abstract": AbstractsFieldName,
    "Keywords": KeywordsFieldName,
    "Section": SectionsFieldName,
    "List": ListsFieldName,
    "Bibliography": BibliographiesFieldName,
    "Equation": EquationsFieldName,
    "Algorithm": AlgorithmsFieldName,
    "Figure": FiguresFieldName,
    "Table": TablesFieldName,
    "Caption": CaptionsFieldName,
    "Header": HeadersFieldName,
    "Footer": FootersFieldName,
    "Footnote": FootnotesFieldName,
}


class CoreRecipe(Recipe):
    def __init__(
        self,
        effdet_publaynet_predictor_path: str = "lp://efficientdet/PubLayNet",
        effdet_mfd_predictor_path: str = "lp://efficientdet/MFD",
        ivila_predictor_path: str = "allenai/ivila-row-layoutlm-finetuned-s2vl-v2",
        bio_roberta_predictor_path: str = "allenai/vila-roberta-large-s2vl-internal",
        svm_word_predictor_path: str = "https://ai2-s2-research-public.s3.us-west-2.amazonaws.com/mmda/models/svm_word_predictor.tar.gz",
        dpi: int = 72,
    ):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.dpi = dpi

        self.logger.info("Instantiating recipe...")
        self.parser = PDFPlumberParser()
        self.rasterizer = PDF2ImageRasterizer()

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.word_predictor = SVMWordPredictor.from_path(svm_word_predictor_path)

        self.effdet_publaynet_predictor = LPBlockPredictor.from_pretrained(effdet_publaynet_predictor_path)
        # self.effdet_mfd_predictor = LPBlockPredictor.from_pretrained(effdet_mfd_predictor_path)
        self.ivila_predictor = IVILATokenClassificationPredictor.from_pretrained(ivila_predictor_path)
        self.bio_roberta_predictor = HFBIOTaggerPredictor.from_pretrained(
            bio_roberta_predictor_path,
            entity_name="tokens",
            context_name="pages",
        )
        self.sent_predictor = PysbdSentencePredictor()
        self.logger.info("Finished instantiating recipe")

    def run(self, pdf: Union[str, Path, Document]) -> Document:
        if isinstance(pdf, str):
            pdf = Path(pdf)
            assert pdf.exists(), f"File {pdf} does not exist."
        assert isinstance(
            pdf, (Document, Path)
        ), f"Unsupported type {type(pdf)} for pdf; should be a Document or a path to a PDF file."
        if isinstance(pdf, Path):
            return self.from_path(str(pdf))
        else:
            raise NotImplementedError("Document input not yet supported.")

    def from_path(self, pdfpath: str) -> Document:
        self.logger.info("Parsing document...")
        doc = self.parser.parse(input_pdf_path=pdfpath)

        self.logger.info("Rasterizing document...")
        images = self.rasterizer.rasterize(input_pdf_path=pdfpath, dpi=self.dpi)
        doc.annotate_images(images=list(images))
        self.rasterizer.attach_images(images=images, doc=doc)
        return self.from_doc(doc=doc)

    def from_doc(self, doc: Document) -> Document:
        self.logger.info("Predicting words...")

        words = self.word_predictor.predict(doc=doc)
        doc.annotate_entity(field_name=WordsFieldName, entities=words)

        self.logger.info("Predicting sentences...")
        sentences = self.sent_predictor.predict(doc=doc)
        doc.annotate_entity(field_name=SentencesFieldName, entities=sentences)

        self.logger.info("Predicting blocks...")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            blocks = self.effdet_publaynet_predictor.predict(doc=doc)
        doc.annotate_entity(field_name=BlocksFieldName, entities=blocks)

        self.logger.info("Predicting vila...")
        vila_entities = self.ivila_predictor.predict(doc=doc)
        doc.annotate_entity(field_name="vila_entities", entities=vila_entities)

        for entity in vila_entities:
            entity.boxes = [
                Box.create_enclosing_box(
                    [b for t in doc.find_by_span(entity, field_name=TokensFieldName) for b in t.boxes]
                )
            ]
            entity.text = make_text(entity=entity, document=doc)
        preds = group_by(entities=vila_entities, metadata_field="label", metadata_values_map=VILA_LABELS_MAP)
        doc.annotate(*preds)
        return doc


if __name__ == "__main__":
    import argparse
    import json

    parser = argparse.ArgumentParser()
    parser.add_argument("--pdf", required=True, type=str, help="Path to PDF file.")
    parser.add_argument("--output", type=str, help="Path to output JSON file.")
    args = parser.parse_args()

    recipe = CoreRecipe()
    doc = recipe.from_path(pdfpath=args.pdf)
    with open(args.output, "w") as f:
        json.dump(doc.to_json(), f, indent=2)



"""

Needs to be imported in this order.

@kylel

"""


from .annotation import Annotation
from .box import Box
from .document import (
    AbstractsFieldName,
    AlgorithmsFieldName,
    AuthorsFieldName,
    BibliographiesFieldName,
    BlocksFieldName,
    CaptionsFieldName,
    Document,
    Prediction,
    EntitiesFieldName,
    EquationsFieldName,
    FiguresFieldName,
    FootersFieldName,
    FootnotesFieldName,
    HeadersFieldName,
    ImagesFieldName,
    KeywordsFieldName,
    ListsFieldName,
    MetadataFieldName,
    PagesFieldName,
    ParagraphsFieldName,
    RelationsFieldName,
    RowsFieldName,
    SectionsFieldName,
    SentencesFieldName,
    SymbolsFieldName,
    TablesFieldName,
    TitlesFieldName,
    TokensFieldName,
    WordsFieldName,
)
from .entity import Entity
from .image import Image
from .indexer import EntityBoxIndexer, EntitySpanIndexer
from .metadata import Metadata
from .span import Span

__all__ = [
    "AbstractsFieldName",
    "AlgorithmsFieldName",
    "Annotation",
    "AuthorsFieldName",
    "BibliographiesFieldName",
    "BlocksFieldName",
    "Box",
    "Prediction",
    "CaptionsFieldName",
    "Document",
    "EntitiesFieldName",
    "Entity",
    "EntityBoxIndexer",
    "EntitySpanIndexer",
    "EquationsFieldName",
    "FiguresFieldName",
    "FootersFieldName",
    "FootnotesFieldName",
    "HeadersFieldName",
    "Image",
    "ImagesFieldName",
    "KeywordsFieldName",
    "KeywordsFieldName",
    "ListsFieldName",
    "Metadata",
    "MetadataFieldName",
    "PagesFieldName",
    "ParagraphsFieldName",
    "RelationsFieldName",
    "RowsFieldName",
    "SectionsFieldName",
    "SentencesFieldName",
    "Span",
    "SymbolsFieldName",
    "TablesFieldName",
    "TitlesFieldName",
    "TokensFieldName",
    "WordsFieldName",
]
